from .azure import login_acr
from .depcheck import checkdep
from .run import promised_run
from .colorc import cc
import sys
import re

re_remote_image_global = re.compile(r"^(.*?)\.azurecr\.io\/(.*)$")
re_remote_image_china = re.compile(r"^(.*?)\.azurecr\.cn\/(.*)$")


def build_docker_image(
    dockerfile: str, build_context: str, image_name: str, no_cache: bool = False
):
    checkdep(["docker"])
    cmd = f'docker build -t "{image_name}" -f "{dockerfile}" "{build_context}"'
    if no_cache:
        cmd += " --no-cache"
    promised_run(cmd)


def push_docker_image(local_image_name: str, remote_image_path: str, tag: str):
    deps = ["docker"]

    match = re_remote_image_global.match(remote_image_path)

    if not match:
        match = re_remote_image_china.match(remote_image_path)

    if match:
        deps.append("az")
        repo_name = match.group(1)
    else:
        cc.error(f"Invalid remote_image_path: {remote_image_path}")
        sys.exit(999)

    checkdep(deps)

    login_acr(repo_name)

    promised_run(f'docker tag "{local_image_name}" "{remote_image_path}:{tag}"')
    promised_run(f'docker push "{remote_image_path}:{tag}"')



import subprocess

from objects.profiles import HelmRepoProfile
from .azure import retry_with_login_az_acr
from .depcheck import checkdep
from .run import run, promised_run, run_get_output


def apply_helm_package(
    dpname: str,
    namespace: str,
    kubecontext: str,
    package: str,
    values: str,
    tag: str,
    repo: HelmRepoProfile | None,
    tag_path: str,
):
    checkdep(["helm", "kubectl"])

    # create namespace if not exists
    current_context = run_get_output("kubectl config current-context")

    def resume_context():
        run(f'kubectl config use-context "{current_context}"')

    promised_run(f'kubectl config use-context "{kubecontext}"', fallback=resume_context)
    proc = run(
        f'kubectl get namespace "{namespace}"',
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    if proc.returncode != 0:
        promised_run(f'kubectl create namespace "{namespace}"', fallback=resume_context)

    if repo is not None:
        # use remote chart
        cmd = f'helm install "{dpname}" "{repo.name}/{package}" --kube-context "{kubecontext}" -n "{namespace}" -f "{values}"'
    else:
        # use local chart
        cmd = f'helm upgrade --kube-context "{kubecontext}" --install "{dpname}" "{package}" -n "{namespace}" -f "{values}"'

    if tag is not None and len(tag) > 0:
        cmd = cmd + f' --set {tag_path}="{tag}"'

    promised_run(cmd, fallback=resume_context)
    resume_context()


def add_repo(name: str, url: str):
    checkdep(["helm"])

    run(f'helm repo add "{name}" "{url}"')
    promised_run(f"helm repo update")



import subprocess
import sys
import functools
from .colorc import cc as colorc
from .depcheck import checkdep
from .run import run


def login_az():
    checkdep(["az"])
    return run("az login").returncode == 0


def retry_with_login_az(func):
    @functools.wraps(func)
    def wrapper(*argv, **kwargs):
        if not func(*argv, **kwargs):
            colorc.warn(f"Failed to calling {func.__name__} for the first time.")
            if not login_az():
                if not func(*argv, **kwargs):
                    colorc.error("Operation failed even after re-loggin.")
                    sys.exit(-3)
        return True

    return wrapper


@retry_with_login_az
def login_acr(acr_name: str):
    checkdep(["az"])
    return run('az acr login -n "' + acr_name + '"').returncode == 0


def retry_with_login_az_acr(name: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*argv, **kwargs):
            if not func(*argv, **kwargs):
                colorc.warn(f"Failed to calling {func.__name__} for the first time.")
                if not login_acr(name):
                    if not func(*argv, **kwargs):
                        colorc.error("Operation failed even after re-loggin.")
                        sys.exit(-3)
            return True

        return wrapper

    return decorator


@retry_with_login_az
def link_aks(subscription: str, resource_group: str, aks: str):
    checkdep(["az", "kubectl"])
    proc = run(f"az account set --subscription {subscription}")
    if proc.returncode != 0:
        return proc.returncode
    proc = run(f"az aks get-credentials --resource-group {resource_group} --name {aks}")
    return proc.returncode == 0



import subprocess
import platform
import sys
from typing import Callable

from .colorc import cc


def run(
    cmd: str, preferred_shell: list[str] = None, stderr=None, stdout=None, cwd=None
):
    ps = preferred_shell

    if ps is None:
        system = platform.system()
        if system == "Windows":
            ps = ["powershell.exe", "-c"]
        elif system == "Linux" or system == "Unix" or system == "Darwin":
            ps = ["/bin/bash", "-c"]
        else:
            ps = []

    ps.append(cmd)

    return subprocess.run(ps, stderr=stderr, stdout=stdout, cwd=cwd)


def promised_run(
    cmd: str,
    preferred_shell: list[str] = None,
    stderr=None,
    stdout=None,
    cwd=None,
    fallback=None,
):
    proc = run(cmd, preferred_shell, stderr, stdout, cwd)
    if proc.returncode != 0:
        if isinstance(fallback, Callable):
            fallback()
        cc.error(f"Failed to run `{cmd}`.")
        sys.exit(999)


def run_get_output(cmd: str, preferred_shell: list[str] = None, cwd=None):
    proc = run(cmd, preferred_shell, subprocess.PIPE, subprocess.PIPE, cwd)
    return proc.stdout.decode().strip()



class bcolors:
    END = "0"
    BOLD = "1"
    ITALIC = "3"
    URL = "4"
    BLINK = "5"
    BLINK2 = "6"
    SELECTED = "7"
    BLACK = "30"
    RED = "31"
    GREEN = "32"
    YELLOW = "33"
    BLUE = "34"
    VIOLET = "35"
    BEIGE = "36"
    WHITE = "37"
    GREY = "90"
    RED2 = "91"
    GREEN2 = "92"
    YELLOW2 = "93"
    BLUE2 = "94"
    VIOLET2 = "95"
    BEIGE2 = "96"
    WHITE2 = "97"
    BGBLACK = "40"
    BGRED = "41"
    BGGREEN = "42"
    BGYELLOW = "43"
    BGBLUE = "44"
    BGVIOLET = "45"
    BGBEIGE = "46"
    BGWHITE = "47"
    BGGREY = "100"
    BGRED2 = "101"
    BGGREEN2 = "102"
    BGYELLOW2 = "103"
    BGBLUE2 = "104"
    BGVIOLET2 = "105"
    BGBEIGE2 = "106"
    BGWHITE2 = "107"

class cc:
    use_verbose: bool = True

    def disable_verbose():
        cc.use_verbose = False

    def error(msg: str):
        print(f"\033[{bcolors.BGRED};{bcolors.BLACK}m ERROR \033[0m {msg}")


    def info(msg: str):
        print(f"\033[{bcolors.BGBLUE};{bcolors.BLACK}m INFO \033[0m {msg}")


    def warn(msg: str):
        print(f"\033[{bcolors.BGYELLOW};{bcolors.BLACK}m WARN \033[0m {msg}")

    def verbose(msg: str):
        if cc.use_verbose:
            print(f"\033[{bcolors.BGBLUE};{bcolors.BLACK}m VERB \033[0m {msg}")
    
    def label(label: str, info: str):
        print(f"\033[{bcolors.BLUE2}m * {label}: \033[0m {info}")



from reprint import output
import subprocess
import sys

from .colorc import bcolors, cc
from .run import run

deps = [
    {
        "name": "docker",
        "checkcmd": "docker --version",
        "url": "https://docs.docker.com/engine/install/",
        "checked": False,
    },
    {
        "name": "az",
        "checkcmd": "az version",
        "url": "https://learn.microsoft.com/en-us/cli/azure/",
        "checked": False,
    },
    {
        "name": "git",
        "checkcmd": "git --version",
        "url": "https://git-scm.com/downloads",
        "checked": False,
    },
    {
        "name": "kubectl",
        "checkcmd": "kubectl --help",
        "url": "https://kubernetes.io/docs/",
        "checked": False,
    },
    {
        "name": "helm",
        "checkcmd": "helm --help",
        "url": "https://helm.sh/docs/intro/install/",
        "checked": False,
    },
]


def checkdep(candidates: list[str] = None):
    if candidates is not None:
        tobechecked = []
        for dep in deps:
            if dep["name"] in candidates and not dep["checked"]:
                tobechecked.append(dep)
    else:
        tobechecked = deps

    failed_count = 0

    if len(tobechecked) > 0:
        cc.verbose(
            f"Performs a dependencies checking: {','.join([e['name'] for e in tobechecked])}"
        )

    with output(output_type="list", initial_len=len(tobechecked)) as output_list:
        run_tag = f"\033[{bcolors.BGYELLOW};{bcolors.BLACK}m WAIT \033[0m "
        ok_tag = f"\033[{bcolors.BGGREEN};{bcolors.BLACK}m  OK  \033[0m "
        fail_tag = f"\033[{bcolors.BGRED};{bcolors.BLACK}m FAIL \033[0m "

        for i, dep in enumerate(tobechecked):
            line = f"{run_tag} Checking \033[{bcolors.BLUE}m{dep['name']}\033[0m"
            output_list[i] = line
            dep["checked"] = True

        for i, dep in enumerate(tobechecked):
            cmd = dep["checkcmd"]
            proc = run(cmd, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
            if proc.returncode == 0:
                output_list[
                    i
                ] = f"{ok_tag} \033[{bcolors.BLUE}m{dep['name']}\033[0m checked."
            else:
                output_list[
                    i
                ] = f"{fail_tag} Installment for \033[{bcolors.BLUE}m{dep['name']}\033[0m, please see \033[{bcolors.BLUE}m{dep['url']}\033[0m"
                failed_count += 1

    if failed_count > 0:
        cc.error("Failed to check the dependencies.")
        sys.exit(999)



import os
import yaml
import sys
from utils.colorc import cc


def null_check(value, name, parent):
    if value is None:
        raise ValueError(f"`{name}` for {parent} could not be empty.")


class DeploymentProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        if isinstance(d, dict):
            if "kubecontexts" in d.keys() and isinstance(d.get("kubecontexts"), list):
                self.kubecontexts = [
                    KubernetesContextProfile(**e) for e in d.get("kubecontexts")
                ]
            else:
                self.kubecontexts = []
            if "docker" in d.keys() and isinstance(d.get("docker"), list):
                self.docker = [DockerImageProfile(**e) for e in d.get("docker")]
            else:
                self.docker = []
            if "helm" in d.keys() and isinstance(d.get("helm"), list):
                self.helm = [HelmPackageProfile(**e) for e in d.get("helm")]
            else:
                self.helm = []
            if "helm_repos" in d.keys() and isinstance(d.get("helm_repos"), list):
                self.repos = [HelmRepoProfile(**e) for e in d.get("helm_repos")]


class KubernetesContextProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        if isinstance(d, dict):
            self.name = d.get("name")
            self.provider = d.get("provider")
            # provider = azure
            self.subscription = d.get("subscription")
            self.resource_group = d.get("resource_group")
            self.aks_name = d.get("aks_name")


class HelmRepoProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        if isinstance(d, dict):
            self.name = d.get("name")
            self.url = d.get("url")


class DockerImageProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        if isinstance(d, dict):
            self.name = d.get("name")
            self.image_name = d.get("image_name")
            self.repository = d.get("repository")
            self.dockerfile = d.get("dockerfile")
            self.context = d.get("context")


class HelmPackageProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        self.name = d.get("name")
        self.namespace = d.get("namespace")
        self.dpname = d.get("dpname")
        self.package = d.get("package")
        self.repo = d.get("repo")
        self.static_tag = d.get("static_tag")
        self.tag_path = d.get("tag_path")
        if "environments" in d.keys() and isinstance(d.get("environments"), list):
            self.environments = [
                HelmEnvironmentProfile(**e) for e in d.get("environments")
            ]
        else:
            self.environments = []
        assert isinstance(self.environments, list)


class HelmEnvironmentProfile:
    def __init__(self, **kwargs) -> None:
        self.from_dict(kwargs)

    def from_dict(self, d: dict) -> None:
        self.name = d.get("name")
        self.values_file = d.get("values_file")
        self.kubecontext = d.get("kubecontext")


def load_profile(file: str):
    if os.path.exists(file) and os.path.isfile(file):
        with open(file, "r", encoding="utf-8") as f:
            d = yaml.load(f, Loader=yaml.FullLoader)
            return DeploymentProfile(**d)
    else:
        cc.error(f"{file} is not a file or not exists.")
        sys.exit(999)
